# Generated by Django 4.2.7 on 2023-12-20 07:27

import pgvector.django
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0023_usersearchmodelconfig"),
    ]

    operations = [
        migrations.AlterField(
            model_name="entry",
            name="embeddings",
            field=pgvector.django.VectorField(),
        ),
    ]
