# Generated by Django 4.2.10 on 2024-04-24 04:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0036_delete_offlinechatprocessorconversationconfig"),
    ]

    operations = [
        migrations.AddField(
            model_name="searchmodelconfig",
            name="bi_encoder_docs_encode_config",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="searchmodelconfig",
            name="bi_encoder_model_config",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="searchmodelconfig",
            name="bi_encoder_query_encode_config",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="searchmodelconfig",
            name="cross_encoder",
            field=models.CharField(default="mixedbread-ai/mxbai-rerank-xsmall-v1", max_length=200),
        ),
    ]
